#-------------------------------------------------------------------------------
# zimport v0.1.5 20250602
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
# refer to https://stackoverflow.com/questions/57651488/how-to-understand-pythons-module-lookup
# refer to https://docs.python.org/3/reference/import.html#the-meta-path
# refer to https://peps.python.org/pep-0302/

import io, os, sys, importlib, time, shutil
import ctypes
import pathlib
import builtins, tokenize

from .main_impl import hook_fileio
from .main_impl import detour
from .util.path import path_exists_native, find, exists
from .util.zip import zipstaties
import zimport.util.zip as ZIP

DBG = False

from . import main_impl
from . import pathfinder
from . import pathfinder_impl

def debug(debug=True) :
    global DBG
    DBG = debug
    main_impl.DBG = debug
    pathfinder.DBG = debug
    pathfinder_impl.DBG = debug
    from .util import zip
    from .util import bootstrap
    from .util import path
    from .util import compile
    bootstrap.DBG = debug
    compile.DBG = debug


########################################

CACHE_DIR_ROOT = None 

def auto_cache_dir() :
    global CACHE_DIR_ROOT
    if "PROJECT_HOME" in os.environ :
        FROM_STRING = "os.environ['PROJECT_HOME']"
        os.environ["PROJECT_HOME"] = os.environ["PROJECT_HOME"].replace('\"', '') # remove '"'
        CACHE_DIR_ROOT = exists(os.environ["PROJECT_HOME"], ['.', '..', 'lib', 'lib/site-packages'], '.cache')
        if CACHE_DIR_ROOT is None : CACHE_DIR_ROOT = os.environ["PROJECT_HOME"] + '/' + ".cache"
    else :
        FROM_STRING = "os.path.dirname(sys.executable)"
        CACHE_DIR_ROOT = exists(os.path.dirname(sys.executable), ['.', '..', 'lib', 'lib/site-packages'], '.cache')
        if CACHE_DIR_ROOT is None : CACHE_DIR_ROOT = os.path.dirname(sys.executable) + '/' + ".cache"

    CACHE_DIR_ROOT = CACHE_DIR_ROOT.replace('\\', '/')
    if not path_exists_native(CACHE_DIR_ROOT) : os.makedirs(CACHE_DIR_ROOT, exist_ok=True)
    print(f"[INF] zimport cache_dir ::: [{CACHE_DIR_ROOT}] from {FROM_STRING}", file=sys.stderr)

def cached_dir(path) : # a/b/c.z/d to library.dir/.cache/a/b/c.z/d
    # if CACHE_DIR_ROOT is None : auto_cache_dir()
    z = path[path.rfind('/'):] # path = c:/a/b/c.z, z = /c.z
    p = None
    if CACHE_DIR_ROOT :
        p = ''.join([CACHE_DIR_ROOT, z])
    else :
        p = ''.join([path, "/../", ".cache", z])
    return p

########################################

PY_VERSION_NUM = f"{sys.version_info.major}{sys.version_info.minor}"
PY_CACHESURFIX = '.cpython-' + PY_VERSION_NUM
#PYPYD_SUFFIXES = [s[0] for s in imp.get_suffixes() if s[2] == imp.C_EXTENSION] # imp deprecated since version 3.4, removed in version 3.12.
PYPYD_SUFFIXES = importlib.machinery.EXTENSION_SUFFIXES
#print(f"PYPYD_SUFFIXES ::: {PYPYD_SUFFIXES}") # ex) PYPYD_SUFFIXES ::: ['.cp311-win_amd64.pyd', '.pyd']

class zimport(object):
    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if hasattr(cls, "_instance"): return cls._instance
        if False : print(f"invoke as singletone(__new__) ::: {cls.__name__}")
        cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        cls = type(self)
        if hasattr(cls, "_init"): return
        if False : print(f"invoke as singletone(__init__) ::: {cls.__name__}")
        cls._init = True
        zimport.__instance__ = self

        print("[INF] zimport installed ...", file=sys.stderr)
        if CACHE_DIR_ROOT is None : auto_cache_dir()
        self.ZIP_REG_NAMES = set()
        self.ZIP_NTRY_INFO = {}
        self.ZIP_STAT_INFO = {}
        self.install_hook()
        self.install_thirdparty_support_pip()
        self.install_importer()


    def install_hook(self):
        builtins.open = hook_fileio(self, "open", True, False, builtins.open)
        os.stat = hook_fileio(self, "stat", True, False, os.stat)
        if os.name == "nt" : os.add_dll_directory = hook_fileio(self, "adll", True, True, os.add_dll_directory)  # for torch, numpy
        if os.name == "nt" : ctypes.WinDLL = hook_fileio(self, "wdll", True, True, ctypes.WinDLL)  # for scipy
        ctypes.CDLL = hook_fileio(self, "cdll", True, True, ctypes.CDLL)
        pathlib.Path.read_text = hook_fileio(self, "read (path)", False, True, pathlib.Path.read_text)  # for sklearn
        pathlib.Path.read_bytes = hook_fileio(self, "bytes(path)", False, True, pathlib.Path.read_bytes)
        tokenize._builtin_open = builtins.open # 20250520 torch/_dynamo/config.py patch
        importlib.machinery.FileFinder.find_spec = detour(self, "FileFinder.find_spec", importlib.machinery.FileFinder.find_spec) # 20250531 torchvision patch
        os.path.exists = detour(self, "os.path.exists", os.path.exists) # 20250531 cv2 patch
        os.listdir = detour(self, "os.listdir", os.listdir) # 202506xx transformers patch

        #os.path.dirname = detour(self, "os.path.dirname", os.path.dirname)  # ...
        #os.path.realpath  = detour(self, "os.path.realpath", os.path.realpath) # ...

        pass

    def install_importer(self):
        from zimport import pathfinder
        sys.path_importer_cache.clear() # very important, if disable occur happen ghost effect, ie, debug mode sucess, normal mode fail
        sys.path_hooks.insert(0, pathfinder.PathFinder)
        pass

    def install_thirdparty_support_pip(self):
        try :
            from zimport import pathfinder
            from pip._vendor.distlib import resources # pip support
            sys.path_importer_cache.clear() # very important, if disable occur happen ghost effect, ie, debug mode sucess, normal mode fail
            resources._finder_registry[pathfinder.PathFinder] = resources.ZipResourceFinder
        except :
            pass
        pass

    ####

    def cached_dir(self, path) : # a/b/c.z/d to library.dir/.cache/a/b/c.z/d
        return cached_dir(path)

    ####

    def getentries(self, zip) : # path must be .z/.zip file path with all unix separator '/' not '\'
        if zip in self.ZIP_NTRY_INFO :
            zent = self.ZIP_NTRY_INFO[zip]
            return zent
        else :
            if DBG: print(f"[INF] new zip-archive {zip} ::: {zip}")
            self.addarchive(zip)
            zent = self.ZIP_NTRY_INFO[zip]
            return zent

    def addarchive(self, zip) : # path must be .z/.zip file path with all unix separator '/' not '\'
        try :
            if zip in self.ZIP_NTRY_INFO :
                print(f"[ERR] already has ntry {zip}", file=sys.stderr)
            if zip in self.ZIP_STAT_INFO :
                print(f"[ERR] already has stat {zip}", file=sys.stderr)
            ntry, stat = ZIP.zipentriesbycase(zip, True, False, True, False)
            self.ZIP_REG_NAMES.add(zip)
            self.ZIP_NTRY_INFO[zip] = ntry
            self.ZIP_STAT_INFO[zip] = stat
            self.addsystempath(zip, ntry.keys())
        except Exception as e :
            print(f"[ERR] addarchive {zip} ::: {e}", file = sys.stderr)

    def fixarchive(self, zip) :
        try:
            ntry, stat = ZIP.zipentriesbycase(zip, True, False, True, False)
            self.ZIP_NTRY_INFO[zip] = ntry
            self.ZIP_STAT_INFO[zip] = stat
        except :
            self.ZIP_NTRY_INFO.pop(zip, None)
            self.ZIP_STAT_INFO.pop(zip, None)

    def addsystempath(self, zip, ent) :
        def path(p): f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
        unq = set()
        lst = [p for p in ent if p.endswith(".pyd")]
        tmp = [p.replace(PYPYD_SUFFIXES[0], '').replace('/', '.') for p in lst]
        for idx in range(len(lst)):
            pyd = lst[idx]  # a/b/c.cp311-win_amd64.pyd'
            dot = tmp[idx]  # a.b.c
            pth = zip + "/" + pyd  # zipped path. ie, library.dir/x.y.z/a/b/c.cp311-win_amd64.pyd
            pth = path(zip + "/" + pyd)  # cached path. ie, library.dir/.cache/x.y.z/a/b/c.cp311-win_amd64.pyd
            unq.add(os.path.dirname(pth))
        lst = [p for p in ent if (p.endswith(".dll") or p.endswith(".so") or (".so." in p) or p.endswith(".dylib") or p.endswith(".exe"))]
        tmp = [p.replace('/', '.') for p in lst]
        for idx in range(len(lst)):
            dll = lst[idx]  # a/b/c.dll'
            dot = tmp[idx]  # a.b.c
            pth = zip + "/" + dll  # zipped path. ie, library.dir/x.y.z/a/b/c.dll
            pth = path(zip + "/" + dll)  # cached path. ie, library.dir/.cache/x.y.z/a/b/c.dll
            unq.add(os.path.dirname(pth))
        if 0 < len(unq):
            for dir in unq:
                if os.name == "nt" : os.add_dll_directory(dir)
                addsyspath(os.path.abspath(dir) if os.name == "nt" else dir)
                if DBG : print("[INF] add PATH [" + dir + "]")

        for dir in [p for p in sys.path if not (p.endswith(".z") or p.endswith(".zip"))]: addsyspath(os.path.abspath(dir) if os.name == "nt" else dir)
        addsyspath(os.path.abspath(os.path.dirname(sys.executable)))  # python.exe installed dir
        addsyspath(os.path.abspath(os.path.dirname(sys.executable) + "/library/bin"))
        if DBG : 
            paths = os.environ["PATH"].split(';' if os.name == "nt" else ':')
            for i in range(len(paths)) :
                print(f"[PATH][{i}] : {paths[i]}")

    def path_maker(self, org_path) :
        return path_maker(self.ZIP_REG_NAMES, org_path)

    def invalidate_caches(self) :
        invalidate_caches()

    def register_hook(self, hookname, org_func, hookfunc) :
        return register_hook(hookname, org_func, hookfunc)

def addsyspath(path) :
    paths = os.environ["PATH"].split(';' if os.name == "nt" else ':')
    if not (path in paths) :
        os.environ["PATH"] += path + (';' if os.name == "nt" else ':')
        if False : print(f"[INF] add PATH : {path}")

cacheofpath = dict()
def path_maker(ziparchive, org_path):
    if org_path in cacheofpath: return cacheofpath[org_path]  # search cache
    abs_path = os.path.abspath(org_path)
    unixpath = abs_path.replace('\\', '/')
    zip_path = ''
    ent_path = ''
    for p in ziparchive:
        if unixpath.startswith(p):  # a.zip/a/b/c.x startswith a.zip
            zip_path = p
            ent_path = unixpath.replace(''.join([p, '/']), '')  # ent_path = unixpath.replace(p + '/', '')
            break
    new_path = '/'.join([cached_dir(zip_path), ent_path])  # new_path = cachedir(zip_path) + '/' + ent_path
    new_path = os.path.abspath(new_path).replace('\\', '/')
    cacheofpath[org_path] = zip_path, ent_path, new_path  # save cache
    return zip_path, ent_path, new_path

def invalidate_caches():
    if False : print(f"[INF] invalidate_caches ...")
    if True : sys.path_importer_cache.clear()
    if False : importlib.invalidate_caches()   # very danger code
    cacheofpath.clear()

######################################## instance of singleton

def getInstance() :
    if zimport.__instance__ is None: zimport()
    return zimport.__instance__

def install() :
    return getInstance()

def uninstall() :
    #sys.path_hooks.insert(0, pathfinder.PathFinder)
    if sys.path_hooks[0].__qualname__  != "PathFinder" : return
    sys.path_hooks.remove(sys.path_hooks[0])
    cls = type(zimport.__instance__)
    delattr(cls, "_instance")
    delattr(cls, "_init")
    zimport.__instance__ = None
    restore_hook()
    sys.path_importer_cache.clear()
    pass

hookforfunc = dict()
def register_hook(hookname, org_func, hookfunc) :
    hookforfunc[hookname] = org_func, hookfunc
    pass

def restore_hook() :
    for k, v in hookforfunc.items() :
        if False : pass
        elif k == "open" :
            builtins.open = v[0]
            tokenize._builtin_open = v[0]
        elif k == "stat" :
            os.stat = v[0]
        elif k == "adll" :
            os.add_dll_directory = v[0]
        elif k == "wdll" :
            ctypes.WinDLL = v[0]
        elif k == "cdll" :
            ctypes.CDLL = v[0]
        elif k == "read (path)" :
            pathlib.Path.read_text = v[0]
        elif k == "bytes(path)" :
            pathlib.Path.read_bytes = v[0]
        elif k == "FileFinder.find_spec" :
            importlib.machinery.FileFinder.find_spec = v[0]
        elif k == "os.path.exists" :
            os.path.exists = v[0]
        elif k == "os.listdir" :
            os.listdir = v[0]
    hookforfunc.clear()
    pass


######################################## cache directory

def zimport_set_cache_dir(path) :
    global CACHE_DIR_ROOT
    if not path_exists_native(path) :
        os.makedirs(path, exist_ok=True)
    CACHE_DIR_ROOT = path
    pass

def zimport_clear_cache() :
    global CACHE_DIR_ROOT
    if not path_exists_native(CACHE_DIR_ROOT) : return
    for f in os.listdir(CACHE_DIR_ROOT):
        try:
            path = os.path.join(CACHE_DIR_ROOT, f)
            shutil.rmtree(path)
        except Exception :
            pass


######################################## manual extract entry to cache directory

def zimport_extract_to_cache(fle, debug=True) : # preload
    zim = getInstance()
    fle = fle.replace('\\','/')
    pth = None
    if zim is None : print("[ERR] try zimport.install() ...", file=sys.stderr); return None
    def copy(p) : f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
    for z in zim.ZIP_NTRY_INFO:
        lst = zim.ZIP_NTRY_INFO[z]
        lst = [z + "/" + p for p in lst if p.endswith(fle)]
        for fle in lst:
            pth = copy(fle)
            if debug :print("[COPY::file] {}".format(pth), file=sys.stderr)
        if 0 < len(lst) :
            break
    if not pth :
        print(f"[COPY::file] cannot found file {fle}", file=sys.stderr)
    return pth

######################################## deprecated function

def precache_directory_deprecated(dir, debug=True) : # preload
    zim = getInstance()
    dir = dir.replace('\\','/')
    pth = None
    if zim is None : print("[ERR] try zimport.install() ...", file=sys.stderr); return None
    def copy(p) : f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
    for z in zim.ZIP_NTRY_INFO:
        lst = zim.ZIP_NTRY_INFO[z]
        lst = [z + "/" + p for p in lst if p.startswith(dir)]
        for fle in lst:
            if fle.endswith('/') : continue # is directory
            out = copy(fle)
            if pth is None : pth = os.path.dirname(out)
            if debug :print("[COPY:::dir] {}".format(out), file=sys.stderr)
        if 0 < len(lst) :
            break
    if not pth :
        print(f"[COPY:::dir] cannot found directory {dir}", file=sys.stderr)
    return pth

def precache_dll_deprecated(dll, debug=True) : # preload pyd, dll, so
    zim = getInstance()
    fle = dll.replace('\\','/')
    pth = None
    if zim is None : print("[ERR] try zimport.install() ...", file=sys.stderr); return None
    def copy(p) : f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
    for z in zim.ZIP_NTRY_INFO:
        lst = zim.ZIP_NTRY_INFO[z]
        lst = [z + "/" + p for p in lst if (p.endswith(".dll") or p.endswith(".so") or (".so." in p) or p.endswith(".dylib") or p.endswith(".pyd") or p.endswith(".exe"))]
        lst = [z + "/" + p for p in lst if p.endswith(fle)]
        for fle in lst:
            pth = copy(fle)
            if debug : print("[COPY:::dll] {}".format(pth), file=sys.stderr)
            if False : ctypes.WinDLL(pth) # preload test
        if 0 < len(lst) :
            break
    if not pth :
        print(f"[COPY:::dir] cannot found dll {dll}", file=sys.stderr)
    return pth






#-------------------------------------------------------------------------------
# zimport v0.1.5 20250602
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys, time, timeit
import io, _io
import zlib, zipfile
import typing # typing added in version 3.5, https://docs.python.org/3/library/typing.html

from importlib import abc
from importlib import _bootstrap
from importlib import _bootstrap_external

class ZipException(ImportError):
    def __init__(self, *args, **kwargs):
        if False : print(f"[EXP] {args[0]}:::{kwargs}", file=sys.stderr)
        pass

class ZipReader(abc.TraversableResources):
    def __init__(self, loader, module):
        _, _, name = module.rpartition('.')
        self.prefix = loader.virt.replace('\\', '/') + name + '/'
        self.archive = loader.real

    def open_resource(self, resource):
        try:
            return super().open_resource(resource)
        except KeyError as exc:
            raise FileNotFoundError(exc.args[0])

    def is_resource(self, path):
        target = self.files().joinpath(path)
        return target.is_file() and target.exists()

    def files(self):
        path = zipfile.Path(self.archive, self.prefix)
        return path

########################################

SIG_STT_ARCHIVE = b'\x50\x4B\x03\x04'
SIG_END_ARCHIVE = b'\x50\x4B\x05\x06'
CENTRAL_DIR_SZE = 22
MAX_COMMENT_LEN = (1 << 16) - 1

def is_ziparchive(p) -> bool :
    return p.endswith(".z") or p.endswith(".zip")

def is_ziparchive_deep(p) -> bool :
    try:
        with io.open(p, "rb") as z:
            byte = z.read(4)
            return True if byte == SIG_STT_ARCHIVE else False
    except : # is directory
        return False

########################################

USEFUCKFILE = False
FUCKINGFILE = dict()

def open(fle) :
    if fle in FUCKINGFILE:
        if False : print(f"[INF] reuse fd {fle}")
        zio = FUCKINGFILE[fle]
    else:
        # zio = _io.open_code(fle)
        zio = io.open(fle, 'rb', buffering = 16384) # io.DEFAULT_BUFFER_SIZE : 8192
        FUCKINGFILE[fle] = zio
    return zio

def zipentries(fle) :
    val = zipentriesbycase(fle, True, False, False, False)
    return val

def zipstaties(fle) :
    val = zipentriesbycase(fle, False, False, True, False)
    return val

def zipentriesbycase(fle : str, gen_entries_by_partname : bool, gen_entries_by_fullname : bool, gen_st_data_by_partname : bool, gen_st_data_by_fullname : bool) :
    try :
        stf = _bootstrap_external._path_stat(fle) # get file stat struct
        std = _bootstrap_external._path_stat(os.path.dirname(fle)) # get directory stat struct
    except :
        raise ZipException(f"can't open Zip file: {fle!r}")

    entries_by_partname = {} # by partname as d/e.txt
    entries_by_fullname = {} # by fullname as a/b/c.z/d/e.txt
    st_data_by_partname = {} # by partname as d/e.txt
    st_data_by_fullname = {} # by fullname as a/b/c.z/d/e.txt

    try:
        fle = fle.replace('\\', '/') # all path must use separator '/'
        zio = open(fle) if USEFUCKFILE else _io.open_code(fle)

        try:
            zio.seek(-CENTRAL_DIR_SZE, 2)
            header_position = zio.tell()
            buffer = zio.read(CENTRAL_DIR_SZE)
        except OSError:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        if len(buffer) != CENTRAL_DIR_SZE:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        if buffer[:4] != SIG_END_ARCHIVE:
            try:
                zio.seek(0, 2)
                file_size = zio.tell()
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            max_comment_start = max(file_size - MAX_COMMENT_LEN - CENTRAL_DIR_SZE, 0)
            try:
                zio.seek(max_comment_start)
                data = zio.read()
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            pos = data.rfind(SIG_END_ARCHIVE)
            if pos < 0:
                raise ZipException(f'not a Zip file: {fle!r}', path=fle)
            buffer = data[pos:pos + CENTRAL_DIR_SZE]
            if len(buffer) != CENTRAL_DIR_SZE:
                raise ZipException(f"corrupt Zip file: {fle!r}", path=fle)
            header_position = file_size - len(data) + pos

        header_size = _unpack_uint32(buffer[12:16])
        header_offset = _unpack_uint32(buffer[16:20])
        if header_position < header_size:
            raise ZipException(f'bad central directory size: {fle!r}', path=fle)
        if header_position < header_offset:
            raise ZipException(f'bad central directory offset: {fle!r}', path=fle)
        header_position -= header_size
        arc_offset = header_position - header_offset
        if arc_offset < 0:
            raise ZipException(f'bad central directory size or offset: {fle!r}', path=fle)

        try: # start of CD (central directory)
            zio.seek(header_position)
        except OSError:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        while True:
            buffer = zio.read(46)
            if len(buffer) < 4:
                raise EOFError('EOF read where not expected')
            # Start of file header
            if buffer[:4] != b'PK\x01\x02':
                break                                # Bad: Central Dir File Header
            if len(buffer) != 46:
                raise EOFError('EOF read where not expected')

            ''''''
            flags        = _unpack_uint16(buffer[8:10])
            compress     = _unpack_uint16(buffer[10:12])
            time         = _unpack_uint16(buffer[12:14])
            date         = _unpack_uint16(buffer[14:16])
            crc          = _unpack_uint32(buffer[16:20])
            data_size    = _unpack_uint32(buffer[20:24])
            file_size    = _unpack_uint32(buffer[24:28])
            name_size    = _unpack_uint16(buffer[28:30])
            extra_size   = _unpack_uint16(buffer[30:32])
            comment_size = _unpack_uint16(buffer[32:34])
            file_offset  = _unpack_uint32(buffer[42:46])
            header_size  = name_size + extra_size + comment_size
            timestamp    = int(datetime(date, time))
            ''''''

            if file_offset > header_offset:
                raise ZipException(f'bad local header offset: {fle!r}', path=fle)
            file_offset += arc_offset

            try:
                name = zio.read(name_size)
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            if len(name) != name_size:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            try:# on windows, calling fseek to skip over the fields we don't use is slower than reading the data
                # because fseek flushes stdio's internal buffers.    See issue #8745.
                if len(zio.read(header_size - name_size)) != header_size - name_size:
                    raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            if flags & 0x800:
                name = name.decode()
            else:
                try:
                    name = name.decode('ascii')
                except UnicodeDecodeError:
                    raise ZipException(f"unicode decode error: {fle!r}", path=fle)

            is_d = name.endswith('/')    # is dir
            path = '/'.join([fle, name]) # dir is always endswith '/'
            _nme = [p for p in name.split('/') if 0 < len(p)][-1] # file/dir name only ::: low performance
            if False : print(name + " ::: " + _nme)

            if gen_entries_by_partname or gen_entries_by_fullname :
                nt = \
                { # entry data
                    "isd" : is_d       , # is dir
                    "pth" : path       , # simple file/dir name only
                    "nme" : _nme       , # simple file/dir name only
                    "met" : compress   , # compression method refer to https://en.wikipedia.org/wiki/ZIP_(file_format)
                    "esz" : data_size  , # encrypt data size
                    "dsz" : file_size  , # decrypt data size
                    "pos" : file_offset, # data position
                    "tme" : timestamp  , # datetime
                    "crc" : crc        , # crc-code
                }
                if gen_entries_by_partname : entries_by_partname[name] = nt # by partname as d/e.txt
                if gen_entries_by_fullname : entries_by_fullname[path] = nt # by fullname as a/b/c.z/d/e.txt

            if gen_st_data_by_partname or gen_st_data_by_fullname :
                #st = copy.deepcopy(std) if is_d else copy.deepcopy(stf)
                s_ = std if is_d else stf
                st = os.stat_result((s_.st_mode,
                                     s_.st_ino,
                                     s_.st_dev,
                                     s_.st_nlink,
                                     s_.st_uid,
                                     s_.st_gid,
                                     file_size,
                                     timestamp,
                                     timestamp,
                                     timestamp,
                                     0, # ? s_.st_blocks,
                                     0, # ? s_.st_blksize,
                                     0, # ? s_.st_rdev,
                                     0, 0, 0, 0, 0))
                if gen_st_data_by_partname : st_data_by_partname[name] = st # by partname as d/e.txt
                if gen_st_data_by_fullname : st_data_by_fullname[path] = st # by fullname as a/b/c.z/d/e.txt
    except OSError:
        raise ZipException(f"can't open Zip file: {fle!r}")
    finally:
        if USEFUCKFILE == False : zio.close()

    ret = []
    cnt = 0
    if gen_entries_by_partname: ret.append(entries_by_partname); cnt+=1
    if gen_entries_by_fullname: ret.append(entries_by_fullname); cnt+=1
    if gen_st_data_by_partname: ret.append(st_data_by_partname); cnt+=1
    if gen_st_data_by_fullname: ret.append(st_data_by_fullname); cnt+=1
    return None if cnt == 0 else ret[0] if cnt == 1 else tuple(ret)

########################################

def getbytes(fle : str, ent : dict):
    try:
        isd = ent["isd"]  # is dir
        nme = ent["nme"]  # simple name
        esz = ent["esz"]  # encrypt data size
        pos = ent["pos"]  # data position
        met = ent["met"]  # compression method
    except OSError:
        raise ZipException(f"invalid entry: {ent!r}", path=fle)

    if isd:
        if False : print(f"cannot decompress dir entry: {nme!r}", file=sys.stderr)
        return

    try:
        fle = fle.replace('\\', '/') # all path must use separator '/'
        zio = open(fle) if USEFUCKFILE else _io.open_code(fle)

        try:
            zio.seek(pos)
        except OSError :
            raise ZipException(f"can't read zip file: {fle!r}", path=fle)

        buf = zio.read(30)
        if len(buf) != 30 :
            raise ZipException(f"eof read where not expected: {fle!r}", path=fle)
        if buf[:4] != SIG_STT_ARCHIVE : # bad: local file header
            raise ZipException(f"bad local file header: {fle!r}", path=fle)

        name_size = _unpack_uint16(buf[26:28])
        extr_size = _unpack_uint16(buf[28:30])
        head_size = 30 + name_size + extr_size
        pos += head_size  # Start of file data

        try:
            zio.seek(pos)
        except OSError :
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

        raw = zio.read(esz)
        if len(raw) != esz:
            raise ZipException(f"data size mismatch : {fle!r}", path=fle)
    except OSError:
        raise ZipException(f"can't open Zip file: {fle!r}")
    finally:
        if USEFUCKFILE == False : zio.close()

    if esz < 0 :
        raise ZipException(f"invalid data size: {esz!r}", path=fle)

    bin = raw if met == 0 else zlib.decompress(raw, -15)
    return bin # refer to https://en.wikipedia.org/wiki/ZIP_(file_format)

########################################

def _pack_uint32(x):
    return (int(x) & 0xFFFFFFFF).to_bytes(4, 'little')

def _unpack_uint32(data):
    assert len(data) == 4
    return int.from_bytes(data, 'little')

def _unpack_uint16(data):
    assert len(data) == 2
    return int.from_bytes(data, 'little')

def datetime(d, t):
    return time.mktime((
        (d >> 9) + 1980,    # bits 9..15: year
        (d >> 5) & 0xF,     # bits 5..8: month
        d & 0x1F,           # bits 0..4: day
        t >> 11,            # bits 11..15: hours
        (t >> 5) & 0x3F,    # bits 8..10: minutes
        (t & 0x1F) * 2,     # bits 0..7: seconds / 2
        -1, -1, -1))

########################################

def testcase1() :
    SETUP_CODE = '''
import os, sys    
from zimport.util.zip import zipentries
'''
    TEST_CODE = '''
list = []
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.additional.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.g2p-en.v2.1.0.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.g2p-ko.v0.9.4.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.mecab.ko.v1.3.7.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.librosa.v0.10.2.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.scipy.v1.14.1.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.numba.v0.60.0.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.tensorboard.v2.17.1.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.torch.v2.0.1+cu118.z")
for p in list :
    zipentries(p)
print('.', end='')
'''
    # timeit.repeat statement
    val = timeit.repeat(setup=SETUP_CODE, stmt=TEST_CODE, repeat=10, number=1)
    print()
    print(f"[INF] min({min(val)}) ~ max({max(val)})")
    # [INF] min(0.13779499999873224) ~ max(0.140073399998073)

def testcase2() :
    path = os.path.dirname(sys.executable) + "/lib/site-packages.additional.z"
    dict = zipentriesbycase(path, True, False, False, False)
    ntry = dict['pydub/audio_segment.py']
    byte = getbytes(path, ntry)
    name = ntry["nme"]
    with open(f"w:/{name}", "wb") as b:
        # Write bytes to file
        b.write(byte)
    print("end of job")

def testcase3() :
    SETUP_CODE = '''
import os, sys
import marshal
from zimport.util.zip import zipentries, zipentriesbycase, getbytes
'''
    TEST_CODE = '''
list = []
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.additional.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.g2p-en.v2.1.0.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.g2p-ko.v0.9.4.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.mecab.ko.v1.3.7.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.librosa.v0.10.2.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.scipy.v1.14.1.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.numba.v0.60.0.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.tensorboard.v2.17.1.z")
list.append(os.path.dirname(sys.executable) + "/lib/site-packages.torch.v2.0.1+cu118.z")
for p in list:
    e1, e2, e3, e4 = zipentriesbycase(p, True, True, True, True)
    for n in e1:
        if not (n.endswith(".py") or n.endswith(".pyc")) : continue
        b = getbytes(p, e1[n])
        # if n.endswith(".py")  : c = compile(b, n, 'exec', dont_inherit=True)
        if n.endswith(".pyc") : m = marshal.loads(b[16:])
    if False : print(f"[INF] completed {p} test ...") 
print('.', end='')
'''
    # timeit.repeat statement
    val = timeit.repeat(setup=SETUP_CODE, stmt=TEST_CODE, repeat=10, number=1)
    print()
    print(f"[INF] min({min(val)}) ~ max({max(val)})")
    # no reuse fd               : [INF] min(1.575370700000348) ~ max(1.646620899999106)
    # ok reuse fd               : [INF] min(0.980815700000675) ~ max(1.028009200002998)
    # ok reuse fd and unmarshal : [INF] min(1.310854899998958) ~ max(1.372702799999388)
    # ok reuse fd and compile   : [INF] min(12.60920729999998) ~ max(13.583235800000693)
if __name__ == "__main__":
    # testcase1()
    # testcase2()
    # testcase3()
    pass





#-------------------------------------------------------------------------------
# zimport v0.1.5 20250602
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
# import _frozen_importlib as _bootstrap
# import _frozen_importlib_external as _bootstrap_external
import sys
from importlib import _bootstrap
from importlib import _bootstrap_external

DBG = False

def modulespec(*args, **kwargs) :
    try :
        ret = _bootstrap.ModuleSpec(*args, **kwargs)
        return ret
    except Exception as e :
        if DBG : print(f"[_bootstrap] modulespec : {args} ::: {e}", file = sys.stderr)
        return None

def spec_from_loader(*args, **kwargs) :
    try :
        ret = _bootstrap.spec_from_loader(*args, **kwargs)
        return ret
    except Exception as e :
        if DBG : print(f"[_bootstrap] spec_from_loader : {args} ::: {e}", file = sys.stderr)
        return None

def exec(*args, **kwargs)  :
    try :
        ret = _bootstrap._exec(*args, **kwargs)
        return ret
    except Exception as e :
        #if DBG : print(f"[_bootstrap] exec : {args} ::: {e}", file = sys.stderr)
        return None

def load_module_shim(*args, **kwargs)  :
    try :
        ret = _bootstrap._load_module_shim(*args, **kwargs)
        return ret
    except Exception as e :
        if DBG : print(f"[_bootstrap] load_module_shim : {args} ::: {e}", file = sys.stderr)
        return None

class LoaderBasics: # _bootstrap_external._LoaderBasics
    pass




from setuptools import setup, find_packages

setup(
    name='zimport',
    version='0.1.0',
    description='zimport is used to load and manage python packages from zip-archives.',
    url='https://github.com/waveware4ai/zimport',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    license='GPL-3.0 license',
    author='14mhz',
    author_email='14mhz@hanmail.net',
)
